/**********************************************************************************
 * Filename: pmon_test.c
 * Note:     this file test kernel module pmon.c which is registered as a char device
 *           at /dev/pmon
 **********************************************************************************/
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>
#define MAX_PMC_NUM (6)
static unsigned int pmc_sel[MAX_PMC_NUM];

static inline unsigned int read_744x_upmc1(void);
static inline unsigned int read_744x_upmc2(void);
static inline unsigned int read_744x_upmc3(void);
static inline unsigned int read_744x_upmc4(void);
static inline unsigned int read_744x_upmc5(void);
static inline unsigned int read_744x_upmc6(void);
static  void show_upmcs(unsigned int* upmc);

int main()
{
  static unsigned int upmc_begin[6],upmc_end[6];
  int i, fd, byteCount, len, n_read;
  unsigned int cycles;
  char* textLine= NULL;
  char item[32], delim[32], name[32];
  int total_pmc= 0;
  FILE* p_cpuinfo;

  /* id CPU to decide how many PMCs we have on this machine */
  p_cpuinfo = fopen("/proc/cpuinfo", "r");
  if(p_cpuinfo == NULL)
    {
      printf("ERR: unable to open  cpuinfo \n");
      return 0;
    }
  while((n_read = getline(&textLine, &len, p_cpuinfo))!=-1)
    {
      sscanf( textLine,"%s%s%s*", item, delim,name);
#ifdef DBG_PMON
      printf("INFO: getline %s\n", textLine);
      printf("INFO:item=%s, delim=%s, name=%s\n", item, delim,name);
#endif
      if(!memcmp(item, "cpu", 4))
	{
	  printf("CPU = %s\n", name);
	  if(!memcmp(name, "744",3))
	    total_pmc = 6;
	  else if(!memcmp(name, "745",3))
	    total_pmc = 6;
	  else if(!memcmp(name, "741",3))	    
	    total_pmc = 4;
	  else
	    {
	      printf("ERR: unsupported CPU %s\n", name);
	      return 0;
	    }
	  printf("CPU %s has %d PMCs\n", name, total_pmc);
	  break;
	}
    }
  if(textLine)
    free(textLine);
  fclose(p_cpuinfo);

  /* FIXME: let usr choose which event for which PMCs based on cpuinfo */
  /* hardcode pmc event number for each pmc */
  printf("Please choose PMC events\n");
  for(i=0;i<total_pmc;i++)
    {
      printf("PMC[%d] event:\t", i);
      if(getline(&textLine, &len, stdin)!=-1)
	{
	  sscanf( textLine,"%d*", &pmc_sel[i]);
	  //	  printf("%d\n", pmc_sel[i]);
	}
    }
  for (i=0;i<total_pmc;i++)
    printf("Monitoring events are PMC[%d]:%d\n", i, pmc_sel[i]);
#if 0
  pmc_sel[0]=1;
  pmc_sel[1]=2;
  pmc_sel[2]=4;
  pmc_sel[3]=1;
  pmc_sel[4]=1;
  pmc_sel[5]=1;
#endif
#if 1
  fd = open("/dev/pmon", O_RDWR);
  if(fd == -1)
    {
      printf("ERR: unable to open device /dev/pmon\n");
      return 0;
    }

  /* Write to pmc selection information to pmon device driver */
  write(fd, pmc_sel, sizeof(pmc_sel));
  for(i=0; i<=10;i++)
    {
      byteCount=read(fd, &cycles, sizeof(int) );
      if(byteCount == -1)
	{
	  printf("ERR: read failed\n");
	  return 0; // can read again
	}
      else if (byteCount < sizeof(cycles))
	{
	  printf("ERR: not read enough data\n");
	  return 0; // can read again
	}
      printf("PMC count = 0x%08X\n", cycles);
    }
  close(fd);
#endif
  show_upmcs(upmc_begin);
  printf("Running my code ..\n\n\n");
  asm volatile("eieio");
  show_upmcs(upmc_end);
  return 0;
}

 void show_upmcs(unsigned int* upmc)
{
  int i;
  upmc[0]= read_744x_upmc1();
  upmc[1]= read_744x_upmc2();
  upmc[2]= read_744x_upmc3();
  upmc[3]= read_744x_upmc4();
  upmc[4]= read_744x_upmc5();
  upmc[5]= read_744x_upmc6();
  for (i=0;i<6;i++)
    printf("UPMC[%d]=0x%08x\n", i,upmc[i]);
}


inline static unsigned int read_744x_upmc1(void)
{
  unsigned int val32;
  asm volatile("mfspr %0, 937" : "=r"(val32));
  return val32;
}
inline static unsigned int read_744x_upmc2(void)
{
  unsigned int val32;
  asm volatile("mfspr %0, 938" : "=r"(val32));
  return val32;
}
inline static unsigned int read_744x_upmc3(void)
{
  unsigned int val32;
  asm volatile("mfspr %0, 941" : "=r"(val32));
  return val32;
}
inline static unsigned int read_744x_upmc4(void)
{
  unsigned int val32;
  asm volatile("mfspr %0, 942" : "=r"(val32));
  return val32;
}
inline static unsigned int read_744x_upmc5(void)
{
  unsigned int val32;
  asm volatile("mfspr %0, 929" : "=r"(val32));
  return val32;
}
inline static unsigned int read_744x_upmc6(void)
{
  unsigned int val32;
  asm volatile("mfspr %0, 930" : "=r"(val32));
  return val32;
}
